"""A module containing a unfied client for interacting with all supported APIs.

Typical usage example:

    cred=User(username="username@domain.com")
    client=UnifiedClient(tenant="tenant.com", cred=cred)
    users=client.get_all_users()

"""
import logging

from azol.clients.oauth_http_client import OAuthHTTPClient
from azol.constants import OAuthResourceIDs
from azol.clients import ArmClient, GraphClient, KeyVaultClient
from azol.models.generic_resource import GenericResource
from azol.resources.rbac_roles import rbac_roles

class UnifiedClient():
    """
        A unified HTTP client that can access all supported HTTP clients from azol
    """

    def __init__( self, tenant, cred, *args, **kwargs ):
        self.arm_client=ArmClient(tenant=tenant, cred=cred)
        if cred.credentialType == "user":
            self.graph_client=GraphClient.from_client(self.arm_client)
        else:
            self.graph_client=GraphClient(tenant=tenant, cred=cred)

    def get_tenants( self ):
        """Get user's tenants. See ARMClient"""
        return self.arm_client.get_tenants()

    def whoami( self ):
        app_displayname=None
        home_tenant=None
        current_tenant=None
        principal_type="unknown"
        name=None
        appid=None
        oid=None
        user=None

        claims=self.arm_client.get_token_claims()

        resource=claims["aud"]
        oid=claims["oid"]
        if claims is None:
            # not authenticated
            return results
        if "app_displayname" in claims.keys():
            app_displayname=claims["app_displayname"]
        if "appid" in claims.keys():
            appid=claims["appid"]
        current_tenant=claims["tid"]

        if "upn" in claims.keys():
            principal_type="user"
            user=claims["upn"]
            name=claims["name"]
            if "idp" in claims.keys():
                home_tenant=claims["idp"]
            else:
                home_tenant=current_tenant
        elif "idp" in claims.keys() and "name" in claims.keys():
            principal_type="user"
            name=claims["name"]
            user=claims["unique_name"]
            home_tenant=claims["idp"]
        else:
            principal_type="app"
            home_tenant=claims["idp"]
        
        results = {
            "principal_type": principal_type,
            "user": user,
            "name": name,
            "client": app_displayname,
            "client_id": appid,
            "home_tenant": home_tenant,
            "current_tenant": current_tenant,
            "resource": resource,
            "object_id": oid
        }
        return results

    def switch_tenant( self, tenant ):
        self.graph_client.switch_tenant(tenant)
        self.arm_client.switch_tenant(tenant)

    def get_api_permissions( self, app ):
        return self.graph_client.get_api_permissions(app)

    def get_tenants(self, ):
        return self.arm_client.get_tenants()

    def fetch_tokens( self ):
        self.graph_client.fetch_token()
        self.arm_client.fetch_token()

    def get_kv_client(self, kv_name):
        return KeyVaultClient.from_client(self.arm_client, kv_name)

    def get_graph_client( self ):
        return self.graph_client

    def get_arm_client( self ):
        return self.arm_client

    def get_management_groups( self, expand=False ):
        """Get management groups. See ArmClient"""
        return self.arm_client.get_management_groups(expand)

    def get_directory_role_assignments(self, object=None):
        return self.graph_client.get_directory_role_assignments(object)

    def get_subscriptions( self, expand=False ):
        """Get subscriptions. See ArmClient"""
        return self.arm_client.get_subscriptions(expand)
    
    def get_all_users( self, expand=False ):
        return self.graph_client.get_all_users(expand)
    



"""A module containing the Access Token credential class"""
from azol.utils import parse_jwt
from azol.credentials.entraid_credential import EntraIdCredential

class AccessToken( EntraIdCredential ):
    '''
        A credential object that may be used when all that is known is a raw access token
    '''

    supportedOAuthFlows = [ "raw_token" ]
    credentialType="unknown"
    default_oauth_flow="raw_token"
    def __init__( self, token, *args, **kwargs ):
        super().__init__( *args, **kwargs)
        self.token = token
        _, body, _ = parse_jwt(token)
        self._client_id=body["appid"]



"""Module that contains functions that are useful when working with Entra ID and Azure"""
import logging
import json
import requests
from azol.utils import string_between
from azol.models.sp_login_init_model import SPLoginModel

def start_azure_portal_login(*args):
    # get a nonce cookie from portal, and have the portal initiate a login flow
    resp=requests.get("https://portal.azure.com/signin/idpRedirect.js/?feature.argsubscriptions=true&feature.showservicehealthalerts=true&feature.prefetchtokens=true&feature.internalgraphapiversion=true&feature.selftoken=true&feature.globalresourcefilter=true&feature.msaljs=true&feature.fetchpolicyforrestypes=true&feature.testcrosscloudpuid=true&feature.useredirecthint=true&feature.usetenanthint=true&idpc=0")

    # Save Portal cookies for later - required when finishing login flow
    portalCookies=resp.cookies

    # Extract portal redirect URL
    portalRedirectUrl=string_between(resp.text, "https://login.microsoftonline.com/organizations/oauth2/v2.0/authorize", "\")", include_start=True)

    init_params=SPLoginModel(
        login_uri=portalRedirectUrl,
        args={
            "portalCookies": portalCookies
        }
    )
    return init_params

def end_azure_portal_login(sp_init_args, code, state, id_token):
    data={
        "state": state,
        "code": code,
        "id_token": id_token
    }

    headers={
        "Content-Type": "application/x-www-form-urlencoded"
    }

    resp=requests.post("https://portal.azure.com/signin/index/", data=data, allow_redirects=False, cookies=sp_init_args["portalCookies"], headers=headers)

    boot_params=string_between(resp.text, "MsPortalImpl.setBootstrapStateAndRedirect(", ");")
    portalbootstrapParams="[" + boot_params + "]"
    params=json.loads(portalbootstrapParams)

    token_data=params[1]

    altRefreshToken=token_data["altRefreshToken"]
    oAuthTokenHeader=token_data["oAuthToken"]["authHeader"]
    access_token=oAuthTokenHeader.split()[1] # the header is "bearer token"
    refresh_token=token_data["refreshToken"]
    selfOAuthToken=token_data["selfOAuthToken"]["authHeader"]
    spaAuthCode=token_data["spaAuthCode"]

    return spaAuthCode


"""Module containing a class for an in-memory token cache"""
from azol.caches.azol_cache import AzolCache

class InMemoryTokenCache( AzolCache ):
    """Class inheriting from AzolCache that provides an in-memory token cache"""



"""Module containing a parent class for an azol cache object"""
import logging
from azol.constants import DEFAULTSCOPE

class AzolCache:
    """Azol parent class for token caching"""

    def __init__( self, ):
        self.tokens={}


    def try_get_token(self, tenant_id, client_id, default_scope, scopes,
                      oauth_resource, username=None):
        """
            Attempt to get a token from the cache that matches the requested tenant_id,
            client_id, scope, and user.

            Args:
                - tenant_id: string - the tenant Id that issued the token
                - client_id: string - the Azure service principal client Id to fetch a token for
                - default_scope: bool - if true, use the default scope for the resource
                - scopes: list - a list of scopes to be issued in the token
                - oauth_resource: string - the OAuth resource issued in the token.
                - username: string - optional. The username that should be issued in the token

            Returns:
                Dictionary containing the scope, username, access and refresh tokens 
        """

        if default_scope:
            scopes=[DEFAULTSCOPE]

        if tenant_id not in self.tokens.keys():
            return None

        if oauth_resource not in self.tokens[tenant_id].keys():
            return None

        if client_id not in self.tokens[tenant_id][oauth_resource].keys():
            return None

        for token_data in self.tokens[tenant_id][oauth_resource][client_id]:

            if token_data["type"] == "delegated":

                if username != token_data["username"]:
                    continue

            # If the token in the cache has all of the scopes that are in the requested scope list,
            # return the token data

            token_found=True

            # Assume the token being reviewed contains all the required scopes. If a scope is found
            # that is not in the list of required scopes, set token_found to False.
            # Otherwise, return token_data after testing all scopes
            for s in scopes:
                if s not in token_data["scopes"]:
                    token_found=False
            if token_found:
                return token_data

    def cache_or_update(self, access_token, tenant_id, client_id, default_scope, scopes,
                        oauth_resource, refresh_token=None, username=None, 
                        ests_cookie=None, ests_persistent_cookie=None):
        """
            Saves an access and refresh token to the cache, or updates an existing entry

             Args:
                - access_token: string - an OAuth2 access token to cache
                - tenant_id: string - the tenant Id that issued the token
                - client_id: string - the Azure service principal client Id to fetch a token for
                - default_scope: bool - if true, use the default scope for the resource
                - scopes: list - a list of scopes to be issued in the token
                - oauth_resource: string - the OAuth resource issued in the token.
                - refresh_token: string - refresh token to cache
                - username: string - optional. The username that should be issued in the token
                - ests_cookie: string - optional. ests cookie to cache
                
            Returns:
                Null
        """
        if default_scope:
            scopes=[DEFAULTSCOPE]
        if tenant_id not in self.tokens.keys():
            self.tokens[ tenant_id ] = {}
        if oauth_resource not in self.tokens[ tenant_id ].keys():
            self.tokens[ tenant_id ][ oauth_resource ] = {}
        if client_id not in self.tokens[ tenant_id ][ oauth_resource ].keys():
            self.tokens[ tenant_id ][ oauth_resource ][ client_id ] = []

        # check to see if a token exists already
        for token_data in self.tokens[tenant_id][oauth_resource][client_id]:
            # in this case, we are trying to write an app token so we dont
            # need to check delegated tokens
            if token_data[ "type" ] == "delegated" and refresh_token is None:
                continue
            if token_data[ "type" ] == "delegated" and refresh_token is not None:
                if token_data[ "username" ] != username:
                    continue

                token_scopes = scopes
                for s in token_data[ "scopes" ]:
                    if s not in token_scopes:
                        continue

                # if we get here, our token matches all attributes in the cached token
                # Update the cached token
                # Note: todo: check to make sure the dictionary actually gets update.
                # offline and cant run the code.

                token_data[ "access_token" ] = access_token
                if token_data["type"] == "delegated":
                    token_data[ "refresh_token" ] = refresh_token
                logging.info("updating token in cache")
                break

            if token_data[ "type" ] == "application" and refresh_token is not None:
                continue
            if token_data[ "type" ] == "application" and refresh_token is None:
                #in this case, we want to check if there is an existing SP token to update:
                token_scopes = scopes
                for s in token_data[ "scopes" ]:
                    if s not in token_scopes:
                        continue
                token_data[ "access_token" ] = access_token
                return

        if refresh_token is None:
            new_token = {
                "type": "application",
                "scopes":scopes,
                "access_token": access_token,
            }
        else:
            new_token = {
                "type": "delegated",
                "scopes":scopes,
                "username":username,
                "access_token": access_token,
                "refresh_token": refresh_token,
                "ests_cookie": ests_cookie,
                "ests_persistent_cookie": ests_persistent_cookie
            }
        self.tokens[ tenant_id ][ oauth_resource ][ client_id ].append(new_token)
        


"""A module containing an azure generic resource model"""
from dataclasses import dataclass
from typing import Optional

@dataclass
class ESTS( object ):
    """
        A dataclass object that contains an ests and ests_persistent cookie
    """
    ests: str
    ests_persistent: str
    